@@ -5,39 +5,36 @@ import json
 import logging
 import re
 
-from sql import Null
-from sql.aggregate import Max
-from sql.conditionals import Case
 from collections import defaultdict
 from itertools import groupby
 
+from sql import Literal, Null
+from sql.aggregate import Max
+from sql.conditionals import Case
+
+from trytond.cache import Cache
 from trytond.config import config
 from trytond.i18n import gettext
+from trytond.model import (
+    DeactivableMixin, EvalEnvironment, ModelSQL, ModelView, Unique, Workflow,
+    fields)
 from trytond.model.exceptions import AccessError, ValidationError
-from ..model import (ModelView, ModelSQL, Workflow, DeactivableMixin, fields,
-    Unique, EvalEnvironment)
-from ..report import Report
-from ..wizard import Wizard, StateView, StateAction, Button
-from ..transaction import Transaction
-from ..cache import Cache
-from ..pool import Pool
-from ..pyson import Bool, Eval, PYSONDecoder
-from ..rpc import RPC
-from ..protocols.jsonrpc import JSONDecoder, JSONEncoder
-from ..tools import is_instance_method, cursor_dict, grouped_slice
-from ..tools.string_ import StringMatcher
+from trytond.pool import Pool
+from trytond.protocols.jsonrpc import JSONDecoder, JSONEncoder
+from trytond.pyson import Bool, Eval, PYSONDecoder
+from trytond.report import Report
+from trytond.rpc import RPC
+from trytond.tools import cursor_dict, grouped_slice, is_instance_method
+from trytond.tools.string_ import StringMatcher
+from trytond.transaction import Transaction
+from trytond.wizard import Button, StateAction, StateView, Wizard
 
-__all__ = [
-    'Model', 'ModelField', 'ModelAccess', 'ModelFieldAccess', 'ModelButton',
-    'ModelButtonRule', 'ModelButtonClick', 'ModelButtonReset',
-    'ModelData', 'PrintModelGraphStart', 'PrintModelGraph', 'ModelGraph',
-    'ModelWorkflowGraph',
-    ]
 logger = logging.getLogger(__name__)
 
 _disable_confirm = config.getboolean('optional',
     'disable_button_confirmations', default=False)
 
+
 class ConditionError(ValidationError):
     pass
 
@@ -62,11 +59,11 @@ class Model(ModelSQL, ModelView):
             },
         depends=['module'])
     module = fields.Char('Module',
-       help="Module in which this model is defined", readonly=True)
+       help="Module in which this model is defined.", readonly=True)
     global_search_p = fields.Boolean('Global Search')
     fields = fields.One2Many('ir.model.field', 'model', 'Fields',
        required=True)
-    _get_name_cache = Cache('ir.model.get_name')
+    _get_names_cache = Cache('ir.model.get_names')
 
     @classmethod
     def __setup__(cls):
@@ -80,6 +77,7 @@ class Model(ModelSQL, ModelView):
         cls.__rpc__.update({
                 'list_models': RPC(),
                 'list_history': RPC(),
+                'get_names': RPC(),
                 'global_search': RPC(),
                 })
 
@@ -101,7 +99,8 @@ class Model(ModelSQL, ModelView):
             cursor.execute(*ir_model.insert(
                     [ir_model.model, ir_model.name, ir_model.info,
                         ir_model.module],
-                    [[model.__name__, model._get_name(), model.__doc__,
+                    [[
+                            model.__name__, model._get_name(), model.__doc__,
                             module_name]]))
             cursor.execute(*ir_model.select(ir_model.id,
                     where=ir_model.model == model.__name__))
@@ -111,6 +110,7 @@ class Model(ModelSQL, ModelView):
                     [ir_model.name, ir_model.info],
                     [model._get_name(), model.__doc__],
                     where=ir_model.id == model_id))
+        cls._get_names_cache.clear()
         return model_id
 
     @classmethod
@@ -149,6 +149,25 @@ class Model(ModelSQL, ModelView):
         return [name for name, model in Pool().iterobject()
             if getattr(model, '_history', False)]
 
+    @classmethod
+    def get_name_items(cls):
+        "Return a list of couple mapping models to names"
+        items = cls._get_names_cache.get('items')
+        if items is None:
+            models = cls.search([])
+            items = [(m.model, m.name) for m in models]
+            cls._get_names_cache.set('items', items)
+        return items
+
+    @classmethod
+    def get_names(cls):
+        "Return a dictionary mapping models to names"
+        dict_ = cls._get_names_cache.get('dict')
+        if dict_ is None:
+            dict_ = dict(cls.get_name_items())
+            cls._get_names_cache.set('dict', dict_)
+        return dict_
+
     @classmethod
     def global_search(cls, text, limit, menu='ir.ui.menu'):
         """
@@ -189,16 +208,7 @@ class Model(ModelSQL, ModelView):
 
     @classmethod
     def get_name(cls, model):
-        name = cls._get_name_cache.get(model)
-        if name is None:
-            models = cls.search([('model', '=', model)], limit=1)
-            if models:
-                model, = models
-                name = model.name
-                cls._get_name_cache.set(model, name)
-            else:
-                name = model
-        return name
+        return cls.get_names().get(model, model)
 
 
 class ModelField(ModelSQL, ModelView):
@@ -237,7 +247,16 @@ class ModelField(ModelSQL, ModelView):
             },
         depends=['module'])
     module = fields.Char('Module',
-       help="Module in which this field is defined")
+       help="Module in which this field is defined.")
+    access = fields.Boolean(
+        "Access",
+        states={
+            'invisible': ~Eval('relation'),
+            },
+        depends=['relation'],
+        help="If checked, the access right on the model of the field "
+        "is also tested against the relation of the field.")
+
     _get_name_cache = Cache('ir.model.field.get_name')
 
     @classmethod
@@ -268,6 +287,7 @@ class ModelField(ModelSQL, ModelView):
                 ir_model_field.relation.as_('relation'),
                 ir_model_field.module.as_('module'),
                 ir_model_field.help.as_('help'),
+                ir_model_field.access.as_('access'),
                 where=ir_model.model == model.__name__))
         model_fields = {f['name']: f for f in cursor_dict(cursor)}
 
@@ -278,26 +298,34 @@ class ModelField(ModelSQL, ModelView):
                 relation = field.relation_name
             else:
                 relation = None
+            access = field_name in model.__access__
 
             if field_name not in model_fields:
                 cursor.execute(*ir_model_field.insert(
                         [ir_model_field.model, ir_model_field.name,
                             ir_model_field.field_description,
                             ir_model_field.ttype, ir_model_field.relation,
-                            ir_model_field.help, ir_model_field.module],
-                        [[model_id, field_name, field.string, field._type,
-                                relation, field.help, module_name]]))
+                            ir_model_field.help, ir_model_field.module,
+                            ir_model_field.access],
+                        [[
+                                model_id, field_name,
+                                field.string,
+                                field._type, relation,
+                                field.help, module_name,
+                                access]]))
             elif (model_fields[field_name]['field_description'] != field.string
                     or model_fields[field_name]['ttype'] != field._type
                     or model_fields[field_name]['relation'] != relation
-                    or model_fields[field_name]['help'] != field.help):
+                    or model_fields[field_name]['help'] != field.help
+                    or model_fields[field_name]['access'] != access):
                 cursor.execute(*ir_model_field.update(
                         [ir_model_field.field_description,
                             ir_model_field.ttype, ir_model_field.relation,
-                            ir_model_field.help],
-                        [field.string, field._type, relation, field.help],
-                        where=ir_model_field.id ==
-                        model_fields[field_name]['id']))
+                            ir_model_field.help, ir_model_field.access],
+                        [field.string, field._type, relation, field.help,
+                            access],
+                        where=(ir_model_field.id
+                            == model_fields[field_name]['id'])))
 
     @classmethod
     def clean(cls):
@@ -397,7 +425,7 @@ class ModelField(ModelSQL, ModelView):
             model = Model.__table__()
             cursor.execute(*model.select(model.id, model.model,
                     where=model.id.in_(model_ids)))
-            id2model = dict(cursor.fetchall())
+            id2model = dict(cursor)
 
             trans_args = []
             for rec in res:
@@ -500,7 +528,7 @@ class ModelAccess(DeactivableMixin, ModelSQL, ModelView):
         user = Transaction().user
         model_access = cls.__table__()
         ir_model = Model.__table__()
-        user_group = UserGroup.__table__()
+        user_group = UserGroup.user_group_all_table()
         group = Group.__table__()
 
         access = {}
@@ -512,6 +540,18 @@ class ModelAccess(DeactivableMixin, ModelSQL, ModelView):
         else:
             return access
 
+        def fill_models(Model, models):
+            if Model.__name__ in models:
+                return
+            models.append(Model.__name__)
+            for field_name in Model.__access__:
+                field = getattr(Model, field_name)
+                fill_models(field.get_target(), models)
+        model2models = defaultdict(list)
+        for model in models:
+            fill_models(pool.get(model), model2models[model])
+
+        all_models = list(set(sum(model2models.values(), [])))
         default = {'read': True, 'write': True, 'create': True, 'delete': True}
         access = dict((m, default) for m in models)
         cursor.execute(*model_access.join(ir_model, 'LEFT',
@@ -522,19 +562,36 @@ class ModelAccess(DeactivableMixin, ModelSQL, ModelView):
                 condition=model_access.group == group.id
                 ).select(
                 ir_model.model,
-                Max(Case((model_access.perm_read == True, 1), else_=0)),
-                Max(Case((model_access.perm_write == True, 1), else_=0)),
-                Max(Case((model_access.perm_create == True, 1), else_=0)),
-                Max(Case((model_access.perm_delete == True, 1), else_=0)),
-                where=ir_model.model.in_(models)
-                & (model_access.active == True)
-                & (((user_group.user == user)
-                        & (group.active == True))
+                Max(Case(
+                        (model_access.perm_read == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (model_access.perm_write == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (model_access.perm_create == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (model_access.perm_delete == Literal(True), 1),
+                        else_=0)),
+                where=ir_model.model.in_(all_models)
+                & (model_access.active == Literal(True))
+                & ((
+                        (user_group.user == user)
+                        & (group.active == Literal(True)))
                     | (model_access.group == Null)),
                 group_by=ir_model.model))
-        access.update(dict(
-                (m, {'read': r, 'write': w, 'create': c, 'delete': d})
-                for m, r, w, c, d in cursor.fetchall()))
+        raw_access = {
+            m: {'read': r, 'write': w, 'create': c, 'delete': d}
+            for m, r, w, c, d in cursor}
+
+        for model in models:
+            access[model] = {
+                perm: max(
+                    (raw_access[m][perm] for m in model2models[model]
+                        if m in raw_access),
+                    default=True)
+                for perm in ['read', 'write', 'create', 'delete']}
         for model, maccess in access.items():
             cls._get_access_cache.set((user, model), maccess)
         return access
@@ -678,7 +735,7 @@ class ModelFieldAccess(DeactivableMixin, ModelSQL, ModelView):
         field_access = cls.__table__()
         ir_model = Model.__table__()
         model_field = ModelField.__table__()
-        user_group = UserGroup.__table__()
+        user_group = UserGroup.user_group_all_table()
         group = Group.__table__()
 
         accesses = {}
@@ -704,17 +761,26 @@ class ModelFieldAccess(DeactivableMixin, ModelSQL, ModelView):
                 ).select(
                 ir_model.model,
                 model_field.name,
-                Max(Case((field_access.perm_read == True , 1), else_=0)),
-                Max(Case((field_access.perm_write == True, 1), else_=0)),
-                Max(Case((field_access.perm_create == True, 1), else_=0)),
-                Max(Case((field_access.perm_delete == True, 1), else_=0)),
+                Max(Case(
+                        (field_access.perm_read == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (field_access.perm_write == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (field_access.perm_create == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (field_access.perm_delete == Literal(True), 1),
+                        else_=0)),
                 where=ir_model.model.in_(models)
-                & (field_access.active == True)
-                & (((user_group.user == user)
-                        & (group.active == True))
+                & (field_access.active == Literal(True))
+                & ((
+                        (user_group.user == user)
+                        & (group.active == Literal(True)))
                     | (field_access.group == Null)),
                 group_by=[ir_model.model, model_field.name]))
-        for m, f, r, w, c, d in cursor.fetchall():
+        for m, f, r, w, c, d in cursor:
             accesses[m][f] = {'read': r, 'write': w, 'create': c, 'delete': d}
         for model, maccesses in accesses.items():
             cls._get_access_cache.set((user, model), maccesses)
@@ -797,7 +863,7 @@ class ModelButton(ModelSQL, ModelView):
             ('id', '!=', Eval('id', -1)),
             ],
         depends=['model', 'id'],
-        help="Button that should reset the rules")
+        help="Button that should reset the rules.")
     reset = fields.Many2Many(
         'ir.model.button-button.reset', 'button', 'button_ruled', "Reset",
         domain=[
@@ -1197,15 +1263,17 @@ class ModelData(ModelSQL, ModelView):
             cursor = Transaction().connection.cursor()
 
             cursor.execute(*table.select(table.model, group_by=[table.model]))
-            models = [m[0] for m in cursor.fetchall()]
+            models = [m for m, in cursor]
             cls._has_model_cache.set(None, models)
         return model in models
 
     @classmethod
-    def get_id(cls, module, fs_id):
+    def get_id(cls, module, fs_id=None):
         """
         Return for an fs_id the corresponding db_id.
         """
+        if fs_id is None:
+            module, fs_id = module.split('.', 1)
         key = (module, fs_id)
         id_ = cls._get_id_cache.get(key)
         if id_ is not None:
@@ -1233,8 +1301,9 @@ class ModelData(ModelSQL, ModelView):
             return dict(json.loads(values, object_hook=JSONDecoder()))
         except ValueError:
             # Migration from 3.2
-            from decimal import Decimal
             import datetime
+
+            from decimal import Decimal
             return eval(values, {
                     'Decimal': Decimal,
                     'datetime': datetime,
@@ -1374,7 +1443,7 @@ class ModelGraph(Report):
 
         for model in models:
             if filter and re.search(filter, model.model):
-                    continue
+                continue
             label = '"{' + model.model + '\\n'
             if model.fields:
                 label += '|'
