@@ -11,6 +11,7 @@ import datetime
 import logging
 import uuid
 import mmap
+import re
 try:
     import secrets
 except ImportError:
@@ -23,7 +24,7 @@ from itertools import groupby
 from operator import attrgetter
 from ast import literal_eval
 
-from sql import Literal
+from sql import Literal, Null
 from sql.functions import CurrentTimestamp
 from sql.conditionals import Coalesce, Case
 from sql.aggregate import Count
@@ -37,31 +38,27 @@ try:
 except ImportError:
     bcrypt = None
 
+from trytond.cache import Cache
+from trytond.config import config
+from trytond.exceptions import LoginException, RateLimitException
 from trytond.exceptions import UserError
 from trytond.i18n import gettext
-from ..model import (
-    ModelView, ModelSQL, Workflow, DeactivableMixin, fields, Unique)
-from ..wizard import Wizard, StateView, Button, StateTransition
-from ..tools import grouped_slice
-from ..transaction import Transaction
-from ..cache import Cache
-from ..pool import Pool
-from ..config import config
-from ..pyson import PYSONEncoder, Eval, Bool
-from ..rpc import RPC
-from ..exceptions import LoginException, RateLimitException
+from trytond.model import (
+    ModelView, ModelSQL, Workflow, DeactivableMixin, fields, Unique,
+    avatar_mixin)
+from trytond.pool import Pool
+from trytond.pyson import PYSONEncoder, Eval, Bool
 from trytond.report import Report, get_email
+from trytond.rpc import RPC
 from trytond.sendmail import sendmail_transactional
-from trytond.url import HOSTNAME
+from trytond.tools import grouped_slice
+from trytond.transaction import Transaction
+from trytond.url import host, http_host
+from trytond.wizard import Wizard, StateView, Button, StateTransition
 
-__all__ = [
-    'User', 'LoginAttempt', 'UserAction', 'UserGroup', 'Warning_',
-    'UserApplication', 'EmailResetPassword',
-    'UserConfigStart', 'UserConfig',
-    ]
 logger = logging.getLogger(__name__)
-_has_password = 'password' in config.get(
-    'session', 'authentications', default='password').split(',')
+_has_password = 'password' in re.split('[,+]', config.get(
+    'session', 'authentications', default='password'))
 
 passlib_path = config.get('password', 'passlib')
 if passlib_path:
@@ -84,17 +81,16 @@ def gen_password(length=8):
 
 
 def _send_email(from_, users, email_func):
-    if from_ is None:
-        from_ = config.get('email', 'from')
+    from_cfg = config.get('email', 'from')
     for user in users:
         if not user.email:
             logger.info("Missing address for '%s' to send email", user.login)
             continue
         msg, title = email_func(user)
-        msg['From'] = from_
+        msg['From'] = from_ or from_cfg
         msg['To'] = user.email
         msg['Subject'] = Header(title, 'utf-8')
-        sendmail_transactional(from_, [user.email], msg)
+        sendmail_transactional(from_cfg, [user.email], msg)
 
 
 class PasswordError(UserError):
@@ -105,7 +101,7 @@ class DeleteError(UserError):
     pass
 
 
-class User(DeactivableMixin, ModelSQL, ModelView):
+class User(avatar_mixin(100, 'login'), DeactivableMixin, ModelSQL, ModelView):
     "User"
     __name__ = "res.user"
     name = fields.Char('Name', select=True)
@@ -134,7 +130,8 @@ class User(DeactivableMixin, ModelSQL, ModelView):
         domain=[('usage', '=', 'menu')], required=True)
     pyson_menu = fields.Function(fields.Char('PySON Menu'), 'get_pyson_menu')
     actions = fields.Many2Many('res.user-ir.action', 'user', 'action',
-        'Actions', help='Actions that will be run at login')
+        'Actions', help='Actions that will be run at login.',
+        size=5)
     groups = fields.Many2Many('res.user-res.group',
        'user', 'group', 'Groups')
     applications = fields.One2Many(
@@ -147,11 +144,13 @@ class User(DeactivableMixin, ModelSQL, ModelView):
             'get_language_direction')
     email = fields.Char('Email')
     status_bar = fields.Function(fields.Char('Status Bar'), 'get_status_bar')
+    avatar_badge_url = fields.Function(
+        fields.Char("Avatar Badge URL"), 'get_avatar_badge_url')
     warnings = fields.One2Many('res.user.warning', 'user', 'Warnings')
     sessions = fields.Function(fields.Integer('Sessions'),
             'get_sessions')
     _get_preferences_cache = Cache('res_user.get_preferences')
-    _get_groups_cache = Cache('res_user.get_groups')
+    _get_groups_cache = Cache('res_user.get_groups', context=False)
     _get_login_cache = Cache('res_user._get_login', context=False)
 
     @classmethod
@@ -182,6 +181,9 @@ class User(DeactivableMixin, ModelSQL, ModelView):
             'pyson_menu',
             'actions',
             'status_bar',
+            'avatar',
+            'avatar_url',
+            'avatar_badge_url',
             'warnings',
             'applications',
         ]
@@ -190,9 +192,13 @@ class User(DeactivableMixin, ModelSQL, ModelView):
             'language_direction',
             'groups',
         ]
+        cls._order.insert(0, ('name', 'ASC'))
 
     @classmethod
     def __register__(cls, module_name):
+        pool = Pool()
+        ModelData = pool.get('ir.model.data')
+        model_data = ModelData.__table__()
         cursor = Transaction().connection.cursor()
         super(User, cls).__register__(module_name)
         table = cls.__table_handler__(module_name)
@@ -211,6 +217,13 @@ class User(DeactivableMixin, ModelSQL, ModelView):
         # Migration from 4.2: Remove required on name
         table.not_null_action('name', action='remove')
 
+        # Migration from 5.6: Set noupdate to admin
+        cursor.execute(*model_data.update(
+                [model_data.noupdate], [True],
+                where=(model_data.model == cls.__name__)
+                & (model_data.module == 'res')
+                & (model_data.fs_id == 'user_admin')))
+
     @staticmethod
     def default_menu():
         pool = Pool()
@@ -223,11 +236,8 @@ class User(DeactivableMixin, ModelSQL, ModelView):
         return None
 
     def get_pyson_menu(self, name):
-        pool = Pool()
-        Action = pool.get('ir.action')
         encoder = PYSONEncoder()
-        return encoder.encode(
-            Action.get_action_values(self.menu.type, [self.menu.id])[0])
+        return encoder.encode(self.menu.get_action_value())
 
     def get_language_direction(self, name):
         pool = Pool()
@@ -240,6 +250,9 @@ class User(DeactivableMixin, ModelSQL, ModelView):
     def get_status_bar(self, name):
         return self.name
 
+    def get_avatar_badge_url(self, name):
+        pass
+
     def get_password(self, name):
         return 'x' * 10
 
@@ -281,9 +294,9 @@ class User(DeactivableMixin, ModelSQL, ModelView):
         for user in users:
             # Use getattr to allow to use non User instances
             for test, message in [
-                    (getattr(user, 'name', ''), 'msg_password_name'),
-                    (getattr(user, 'login', ''), 'msg_password_login'),
-                    (getattr(user, 'email', ''), 'msg_password_email'),
+                    (getattr(user, 'name', ''), 'res.msg_password_name'),
+                    (getattr(user, 'login', ''), 'res.msg_password_login'),
+                    (getattr(user, 'email', ''), 'res.msg_password_email'),
                     ]:
                 if test and password.lower() == test.lower():
                     raise PasswordError(gettext(message))
@@ -370,10 +383,12 @@ class User(DeactivableMixin, ModelSQL, ModelView):
     def write(cls, users, values, *args):
         pool = Pool()
         Session = pool.get('ir.session')
+        UserDevice = pool.get('res.user.device')
 
         actions = iter((users, values) + args)
         all_users = []
         session_to_clear = []
+        users_to_clear = []
         args = []
         for users, values in zip(actions, actions):
             all_users += users
@@ -381,17 +396,18 @@ class User(DeactivableMixin, ModelSQL, ModelView):
 
             if 'password' in values:
                 session_to_clear += users
+                users_to_clear += [u.login for u in users]
 
         super(User, cls).write(*args)
 
         Session.clear(session_to_clear)
+        UserDevice.clear(users_to_clear)
 
         # Clean cursor cache as it could be filled by domain_get
         for cache in Transaction().cache.values():
             if cls.__name__ in cache:
                 for user in all_users:
-                    if user.id in cache[cls.__name__]:
-                        cache[cls.__name__][user.id].clear()
+                    cache[cls.__name__].pop(user.id, None)
         # Restart the cache for domain_get method
         pool = Pool()
         pool.get('ir.rule')._domain_get_cache.clear()
@@ -466,14 +482,14 @@ class User(DeactivableMixin, ModelSQL, ModelView):
                     else:
                         res['language'] = Config.get_language()
                 else:
-                    res[field] = None
                     if getattr(user, field):
                         res[field] = getattr(user, field).id
                         res[field + '.rec_name'] = \
                             getattr(user, field).rec_name
             elif cls._fields[field]._type in ('one2many', 'many2many'):
                 res[field] = [x.id for x in getattr(user, field)]
-                if field == 'actions' and user.login == 'admin':
+                admin_id = ModelData.get_id('res.user_admin')
+                if field == 'actions' and user.id == admin_id:
                     config_wizard_id = ModelData.get_id('ir',
                         'act_module_config_wizard')
                     action_id = Action.get_action_id(config_wizard_id)
@@ -588,15 +604,22 @@ class User(DeactivableMixin, ModelSQL, ModelView):
         return res
 
     @classmethod
-    def get_groups(cls):
+    def get_groups(cls, name=None):
         '''
-        Return a list of group ids for the user
+        Return a list of all group ids for the user
         '''
         user = Transaction().user
         groups = cls._get_groups_cache.get(user)
         if groups is not None:
             return groups
-        groups = cls.read([Transaction().user], ['groups'])[0]['groups']
+        pool = Pool()
+        UserGroup = pool.get('res.user-res.group')
+        cursor = Transaction().connection.cursor()
+        user_group = UserGroup.user_group_all_table()
+        cursor.execute(*user_group.select(
+                user_group.group,
+                where=user_group.user == user))
+        groups = [g for g, in cursor]
         cls._get_groups_cache.set(user, groups)
         return groups
 
@@ -612,7 +635,8 @@ class User(DeactivableMixin, ModelSQL, ModelView):
                     (table.password_reset_expire > CurrentTimestamp(),
                         table.password_reset),
                     else_=None),
-                where=(table.login == login) & (table.active == True)))
+                where=(table.login == login)
+                & (table.active == Literal(True))))
         result = cursor.fetchone() or (None, None, None)
         cls._get_login_cache.set(login, result)
         return result
@@ -622,34 +646,43 @@ class User(DeactivableMixin, ModelSQL, ModelView):
         '''
         Return user id if password matches
         '''
-        LoginAttempt = Pool().get('res.user.login.attempt')
+        pool = Pool()
+        LoginAttempt = pool.get('res.user.login.attempt')
+        UserDevice = pool.get('res.user.device')
+
         count_ip = LoginAttempt.count_ip()
         if count_ip > config.getint(
                 'session', 'max_attempt_ip_network', default=300):
             # Do not add attempt as the goal is to prevent flooding
             raise RateLimitException()
-        count = LoginAttempt.count(login)
+        device_cookie = UserDevice.get_valid_cookie(
+            login, parameters.get('device_cookie'))
+        count = LoginAttempt.count(login, device_cookie)
         if count > config.getint('session', 'max_attempt', default=5):
-            LoginAttempt.add(login)
+            LoginAttempt.add(login, device_cookie)
             raise RateLimitException()
-        Transaction().atexit(time.sleep, 2 ** count - 1)
-        for method in config.get(
+        Transaction().atexit(time.sleep, random.randint(0, 2 ** count - 1))
+        for methods in config.get(
                 'session', 'authentications', default='password').split(','):
-            try:
-                func = getattr(cls, '_login_%s' % method)
-            except AttributeError:
-                logger.info('Missing login method: %s', method)
-                continue
-            user_id = func(login, parameters)
-            if user_id:
-                LoginAttempt.remove(login)
-                return user_id
-        LoginAttempt.add(login)
+            user_ids = set()
+            for method in methods.split('+'):
+                try:
+                    func = getattr(cls, '_login_%s' % method)
+                except AttributeError:
+                    logger.info('Missing login method: %s', method)
+                    break
+                user_ids.add(func(login, parameters))
+                if len(user_ids) != 1 or not all(user_ids):
+                    break
+            if len(user_ids) == 1 and all(user_ids):
+                LoginAttempt.remove(login, device_cookie)
+                return user_ids.pop()
+        LoginAttempt.add(login, device_cookie)
 
     @classmethod
     def _login_password(cls, login, parameters):
         if 'password' not in parameters:
-            msg = cls.fields_get(['password'])['password']['string']
+            msg = gettext('res.msg_user_password', login=login)
             raise LoginException('password', msg, type='password')
         user_id, password_hash, password_reset = cls._get_login(login)
         if user_id and password_hash:
@@ -740,6 +773,7 @@ class LoginAttempt(ModelSQL):
     """
     __name__ = 'res.user.login.attempt'
     login = fields.Char('Login', size=512)
+    device_cookie = fields.Char("Device Cookie")
     ip_address = fields.Char("IP Address")
     ip_network = fields.Char("IP Network")
 
@@ -771,7 +805,7 @@ class LoginAttempt(ModelSQL):
 
     @classmethod
     @_login_size
-    def add(cls, login):
+    def add(cls, login, device_cookie=None):
         cursor = Transaction().connection.cursor()
         table = cls.__table__()
         cursor.execute(*table.delete(where=table.create_date < cls.delay()))
@@ -779,24 +813,29 @@ class LoginAttempt(ModelSQL):
         ip_address, ip_network = cls.ipaddress()
         cls.create([{
                     'login': login,
+                    'device_cookie': device_cookie,
                     'ip_address': str(ip_address),
                     'ip_network': str(ip_network),
                     }])
 
     @classmethod
     @_login_size
-    def remove(cls, login):
+    def remove(cls, login, device_cookie=None):
         cursor = Transaction().connection.cursor()
         table = cls.__table__()
-        cursor.execute(*table.delete(where=table.login == login))
+        cursor.execute(*table.delete(
+                where=(table.login == login)
+                & (table.device_cookie == device_cookie)
+                ))
 
     @classmethod
     @_login_size
-    def count(cls, login):
+    def count(cls, login, device_cookie=None):
         cursor = Transaction().connection.cursor()
         table = cls.__table__()
         cursor.execute(*table.select(Count(Literal('*')),
                 where=(table.login == login)
+                & (table.device_cookie == device_cookie)
                 & (table.create_date >= cls.delay())))
         return cursor.fetchone()[0]
 
@@ -813,6 +852,62 @@ class LoginAttempt(ModelSQL):
     del _login_size
 
 
+class UserDevice(ModelSQL):
+    "User Device"
+    __name__ = 'res.user.device'
+
+    login = fields.Char("Login", required=True)
+    cookie = fields.Char("Cookie", readonly=True, required=True)
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls.__rpc__.update({
+                'renew': RPC(readonly=False),
+                })
+
+    @classmethod
+    def get_valid_cookie(cls, login, cookie):
+        try:
+            device, = cls.search([
+                    ('login', '=', login),
+                    ('cookie', '=', cookie),
+                    ], limit=1)
+        except ValueError:
+            return None
+
+        return device.cookie
+
+    @classmethod
+    def renew(cls, current_cookie):
+        pool = Pool()
+        User = pool.get('res.user')
+
+        user = User(Transaction().user)
+        new_cookie = uuid.uuid4().hex
+        current_devices = cls.search([
+                    ('login', '=', user.login),
+                    ('cookie', '=', current_cookie),
+                    ])
+        if current_devices:
+            cls.write(current_devices, {
+                    'cookie': new_cookie
+                    })
+        else:
+            cls.create([{
+                        'login': user.login,
+                        'cookie': new_cookie,
+                        }])
+        return new_cookie
+
+    @classmethod
+    def clear(cls, logins):
+        for sub_logins in grouped_slice(logins):
+            cls.delete(cls.search([
+                        ('login', 'in', list(sub_logins)),
+                        ]))
+
+
 class UserAction(ModelSQL):
     'User - Action'
     __name__ = 'res.user-ir.action'
@@ -852,6 +947,68 @@ class UserGroup(ModelSQL):
     group = fields.Many2One('res.group', 'Group', ondelete='CASCADE',
             select=True, required=True)
 
+    @classmethod
+    def create(cls, vlist):
+        records = super().create(vlist)
+        pool = Pool()
+        # Restart the cache on the domain_get method
+        pool.get('ir.rule')._domain_get_cache.clear()
+        # Restart the cache for get_groups
+        pool.get('res.user')._get_groups_cache.clear()
+        # Restart the cache for get_preferences
+        pool.get('res.user')._get_preferences_cache.clear()
+        # Restart the cache for model access and view
+        pool.get('ir.model.access')._get_access_cache.clear()
+        pool.get('ir.model.field.access')._get_access_cache.clear()
+        ModelView._fields_view_get_cache.clear()
+        return records
+
+    @classmethod
+    def write(cls, groups, values, *args):
+        super().write(groups, values, *args)
+        pool = Pool()
+        # Restart the cache on the domain_get method
+        pool.get('ir.rule')._domain_get_cache.clear()
+        # Restart the cache for get_groups
+        pool.get('res.user')._get_groups_cache.clear()
+        # Restart the cache for get_preferences
+        pool.get('res.user')._get_preferences_cache.clear()
+        # Restart the cache for model access and view
+        pool.get('ir.model.access')._get_access_cache.clear()
+        pool.get('ir.model.field.access')._get_access_cache.clear()
+        ModelView._fields_view_get_cache.clear()
+
+    @classmethod
+    def delete(cls, groups):
+        super().delete(groups)
+        pool = Pool()
+        # Restart the cache on the domain_get method
+        pool.get('ir.rule')._domain_get_cache.clear()
+        # Restart the cache for get_groups
+        pool.get('res.user')._get_groups_cache.clear()
+        # Restart the cache for get_preferences
+        pool.get('res.user')._get_preferences_cache.clear()
+        # Restart the cache for model access and view
+        pool.get('ir.model.access')._get_access_cache.clear()
+        pool.get('ir.model.field.access')._get_access_cache.clear()
+        ModelView._fields_view_get_cache.clear()
+
+    @classmethod
+    def user_group_all_table(cls):
+        pool = Pool()
+        Group = pool.get('res.group')
+        user_group = cls.__table__()
+        group_parents = Group.group_parent_all_cte()
+
+        return (user_group
+            .join(group_parents,
+                condition=user_group.group == group_parents.id)
+            .select(
+                user_group.user.as_('user'),
+                group_parents.parent.as_('group'),
+                where=group_parents.parent != Null,
+                with_=group_parents))
+
 
 class Warning_(ModelSQL, ModelView):
     'User Warning'
@@ -863,8 +1020,10 @@ class Warning_(ModelSQL, ModelView):
 
     @classmethod
     def check(cls, warning_name):
-        user = Transaction().user
-        if not user:
+        transaction = Transaction()
+        user = transaction.user
+        context = transaction.context
+        if not user or context.get('_skip_warnings'):
             return False
         warnings = cls.search([
             ('user', '=', user),
@@ -982,12 +1141,13 @@ class EmailResetPassword(Report):
     __name__ = 'res.user.email_reset_password'
 
     @classmethod
-    def get_context(cls, records, data):
+    def get_context(cls, records, header, data):
         pool = Pool()
         Lang = pool.get('ir.lang')
-        context = super(EmailResetPassword, cls).get_context(records, data)
+        context = super().get_context(records, header, data)
         lang = Lang.get()
-        context['hostname'] = HOSTNAME
+        context['host'] = host()
+        context['http_host'] = http_host()
         context['database'] = Transaction().database.name
         context['expire'] = lang.strftime(
             records[0].password_reset_expire,
